/*
 *    snake, a videogame derived from Q-Basic nibbles
 *    Copyright (C) 1999-2000 Cyberandro J.K. Starrik
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "modules.h"
#include <string.h>		/* strings */
#include <stdlib.h>		/* malloc */
#include <errno.h>

#undef DEBUG

/* WEIGHTS */
#define MAXSCORE		100

#define SNAKE_DISTANCE_WEIGHT(x)  (((brain *)x->brain)->snake_distance_weight)
#define FRUIT_DISTANCE_WEIGHT(x)  (((brain *)x->brain)->fruit_distance_weight)
#define RANDOM_WEIGHT(x)	  (((brain *)x->brain)->random_weight)
#define INERTIA_WEIGHT(x)	  (((brain *)x->brain)->inertia_weight)
#define SNAKE_PRESENCE_WEIGHT(x)  (((brain *)x->brain)->snake_presence_weight)
#define FREE_SPACE_WEIGHT(x)	  (((brain *)x->brain)->free_space_weight)
#define DEADSNAKE_PROXIMITY_WEIGHT(x)	  (((brain *)x->brain)->deadsnake_proximity_weight)

/* local data and definitions */
#ifdef DEBUG
static char Direction[4] = { 'l', 'u', 'r', 'd' };
#endif

typedef struct t_scored_dir
{
  int dir;
  float score;
}
scored_dir;

typedef struct t_brain
{
  float snake_distance_weight, fruit_distance_weight, random_weight,
    inertia_weight, snake_presence_weight, free_space_weight,
    deadsnake_proximity_weight;
  int scored_dir[4];
}
brain;

/* private functions */
static void play (snake * s);
static int parse_config (void);
static int distance (cell * from, cell * to, int dir);
static void score_by_random (snake *, scored_dir **);
static void score_by_inertia (snake *, scored_dir **);
static void score_by_snake_distance (snake *, scored_dir **);
static void score_by_free_space (snake *, scored_dir **);
static void score_by_snake_presence (snake *, fruit *, scored_dir **);
static void score_by_fruit_distance (snake *, fruit *, scored_dir **);
static void score_by_deadsnake_proximity (snake *, scored_dir **);
static void sort_direction_by_score (snake *, scored_dir **);
static brain *make_brain (float, float, float, float, float, float, float);
#ifdef DEBUG
static void dump_scores (snake *, scored_dir **);
#endif

/* public functions */
int ai_module_init (void);

/* module initialization function */
int
ai_module_init ()
{
  if (!parse_config ())
    return 1;

  return 0;
}

/* parse configuration file */
static int
parse_config ()
{
  snake *mysnake;
  char buf[256];
  int ret;
  char format[] = "ai %f %f %f %f %f %f %f\n";
  float w_snake_distance, w_fruit_distance, w_random,
    w_inertia, w_snake_presence, w_free_space, w_deadsnake;

  /* this file is used by every modules ... */
  rewind (conf);

  while (fgets (buf, 255, conf))
    {
      int playernum;

      /* read record */
      ret = sscanf (buf, format,
		    &w_snake_distance,
		    &w_fruit_distance, &w_random,
		    &w_inertia, &w_snake_presence,
		    &w_free_space, &w_deadsnake);

      /* malformed config line (probably does not belong to us) */
      if (ret < 6)
	continue;

#ifdef DEBUG
      fprintf (stderr, "%s", buf);
      fprintf (stderr, "sscanf returned %i\n", ret);
      fprintf (stderr, "%f, %f, %f, %f, %f, %f %f\n",
	       w_snake_distance,
	       w_fruit_distance, w_random,
	       w_inertia, w_snake_presence, w_free_space, w_deadsnake);
#endif

      /* make a new player */
      mysnake = snake_make ();

      /* set player function */
      mysnake->play = &play;

      /* make snake's brain */
      mysnake->brain = (brain *) make_brain (w_snake_distance,
					     w_fruit_distance, w_random,
					     w_inertia, w_snake_presence,
					     w_free_space, w_deadsnake);

      /* register new player */
      playernum = register_player (mysnake);

    }

#ifdef DEBUG
  fprintf (stderr, "snake_distance_weight=%f\n", w_snake_distance);
  fprintf (stderr, "fruit_distance_weight=%f\n", w_fruit_distance);
  fprintf (stderr, "random_weight=%f\n", w_random);
  fprintf (stderr, "inertia_weight=%f\n", w_inertia);
  fprintf (stderr, "snake_presence_weight=%f\n", w_snake_presence);
  fprintf (stderr, "free_space_weight=%f\n", w_free_space);
  fprintf (stderr, "deadsnake_proximity_weight=%f\n", w_deadsnake);
#endif

  return 1;
}

/* score each direction based on given weights. pick the best */
static void
play (snake * s)
{
  cell *headcell;
  cell tmpcell;
  int i;
  scored_dir *scored_dir[4];

  headcell = &(s->head->cell);

  /* init scored_dir array */
  for (i = 0; i < 4; i++)
    {
      scored_dir[i] = malloc (sizeof (scored_dir));
      scored_dir[i]->dir = i;
      scored_dir[i]->score = 0;
    }


  /* score by snake distance */
  score_by_snake_distance (s, scored_dir);

  /* score by space occupation */
  score_by_free_space (s, scored_dir);

  /* score by fruit distance */
  score_by_fruit_distance (s, &frt, scored_dir);

  /* score by deadsnake proximity */
  score_by_deadsnake_proximity (s, scored_dir);

  /* score by snake presence */
  score_by_snake_presence (s, &frt, scored_dir);

  /* score by random ... */
  score_by_random (s, scored_dir);

  /* score by inertia */
  score_by_inertia (s, scored_dir);

#ifdef DEBUG
  /* dump scored (do it after sorting if you like)... */
  dump_scores (s, scored_dir);
#endif

  /* sort directions by score */
  sort_direction_by_score (s, scored_dir);

  // for (i=0; i<200000; i++); /* just waste some time... */

  /* find first free direction (in order of score) */
  for (i = 0; i < 4; i++)
    {
      memcpy (&tmpcell, headcell, sizeof (cell));
      move_cell (&tmpcell, scored_dir[i]->dir, 1);

      if (!IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	{
	  s->dir = scored_dir[i]->dir;
	  break;
	}
    }

  /* if no free directions are found, i will be == 4
   * which is the special 'STOP' direction
   */


  /* finish scored_dir array */
  for (i = 0; i < 4; i++)
    {
      free (scored_dir[i]);
    }

}

/* return distance to ortogonal position between ``from'' and ``to''
 * going in given direction ``dir''
 */
static int
distance (cell * from, cell * to, int dir)
{
  int dist;

  switch (dir)
    {
    case UP:
      dist = from->y - to->y;
      if (dist < 0)
	dist = ARENA_HEIGHT - abs (dist);
      break;
    case DOWN:
      dist = to->y - from->y;
      if (dist < 0)
	dist = ARENA_HEIGHT - abs (dist);
      break;
    case LEFT:
      dist = from->x - to->x;
      if (dist < 0)
	dist = ARENA_WIDTH - abs (dist);
      break;
    case RIGHT:
      dist = to->x - from->x;
      if (dist < 0)
	dist = ARENA_WIDTH - abs (dist);
      break;
    }

#ifdef DEBUG_DISTANCE
  fprintf (stderr, "(%i,%i) <-> (%i,%i) %c %i\n",
	   from->y, from->x, to->y, to->x, Direction[dir], dist);
#endif

  return dist;
}

/* random scoring */
static void
score_by_random (snake * s, scored_dir ** sdir)
{
  int i;
  float weight = RANDOM_WEIGHT (s);

  //i = rand_between(0,3);
  for (i = 0; i < 4; i++)
    sdir[i]->score += rand_between (0, MAXSCORE) * weight;
}

/* score by inertia */
static void
score_by_inertia (snake * s, scored_dir ** sdir)
{
  float weight = INERTIA_WEIGHT (s);
  int *dir;

  dir = &(s->dir);

  sdir[*dir]->score += MAXSCORE * weight;
}

/* score by snake distance */
static void
score_by_snake_distance (snake * s, scored_dir ** dir)
{
  int dist;
  float maxdist;
  cell *from;
  cell tmpcell;
  float weight = SNAKE_DISTANCE_WEIGHT (s);

  from = &(s->head->cell);

  /* horizontal distances */
  maxdist = ARENA_WIDTH;

  /* right */
  dist = 0;
  memcpy (&tmpcell, from, sizeof (cell));
  do
    {
      move_cell (&tmpcell, RIGHT, 1);
      dist++;
      if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	break;
    }
  while (tmpcell.x != from->x);
  dir[RIGHT]->score += (dist / maxdist) * MAXSCORE * weight;

  /* left */
  dist = 0;
  memcpy (&tmpcell, from, sizeof (cell));
  do
    {
      move_cell (&tmpcell, LEFT, 1);
      dist++;
      if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	break;
    }
  while (tmpcell.x != from->x);
  dir[LEFT]->score += (dist / maxdist) * MAXSCORE * weight;


  /* vertical distances */
  maxdist = ARENA_HEIGHT;

  /* up */
  dist = 0;
  memcpy (&tmpcell, from, sizeof (cell));
  do
    {
      move_cell (&tmpcell, UP, 1);
      dist++;
      if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	break;
    }
  while (tmpcell.y != from->y);
  dir[UP]->score += (dist / maxdist) * MAXSCORE * weight;

  /* down */
  dist = 0;
  memcpy (&tmpcell, from, sizeof (cell));
  do
    {
      move_cell (&tmpcell, DOWN, 1);
      dist++;
      if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	break;
    }
  while (tmpcell.y != from->y);
  dir[DOWN]->score += (dist / maxdist) * MAXSCORE * weight;
}

/* score by free space */
static void
score_by_free_space (snake * s, scored_dir ** dir)
{
  int square_side = 3;		// be odd!
  cell tmpcell;
  cell *from;
  int half_side;
  int c, d;
  int freespace;
  int maxspace;
  int weight = FREE_SPACE_WEIGHT (s);

  from = &(s->head->cell);

  half_side = (square_side / 2);
  maxspace = square_side * square_side;

  /* right */
  freespace = maxspace;
  tmpcell.x = from->x;
  for (move_cell (&tmpcell, RIGHT, 1), c = 0;
       c < square_side; c++, move_cell (&tmpcell, RIGHT, 1))
    {
      tmpcell.y = from->y;
      for (move_cell (&tmpcell, UP, 1), d = 0;
	   d < square_side; d++, move_cell (&tmpcell, UP, 1))
	{
	  if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	    {
	      freespace--;
	    }
	}
    }
  dir[RIGHT]->score += freespace * weight;

  /* left */
  freespace = maxspace;
  tmpcell.y = from->y;
  for (move_cell (&tmpcell, UP, half_side), c = 0;
       c < square_side; c++, move_cell (&tmpcell, DOWN, 1))
    {
      tmpcell.x = from->x;
      for (move_cell (&tmpcell, LEFT, square_side), d = 0;
	   d < square_side; d++, move_cell (&tmpcell, RIGHT, 1))
	{
	  if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	    freespace--;
	}
    }
  dir[LEFT]->score += freespace * weight;

  /* up */
  freespace = maxspace;
  tmpcell.x = from->x;
  for (move_cell (&tmpcell, LEFT, half_side), c = 0;
       c < square_side; c++, move_cell (&tmpcell, RIGHT, 1))
    {
      tmpcell.y = from->y;
      for (move_cell (&tmpcell, UP, 1), d = 0;
	   d < square_side; d++, move_cell (&tmpcell, UP, 1))
	{
	  if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	    freespace--;
	}
    }
  dir[UP]->score += freespace * weight;

  /* down */
  freespace = maxspace;
  tmpcell.y = from->y;
  for (move_cell (&tmpcell, DOWN, 1), c = 0;
       c < square_side; c++, move_cell (&tmpcell, DOWN, 1))
    {
      tmpcell.x = from->x;
      for (move_cell (&tmpcell, LEFT, half_side), d = 0;
	   d < square_side; d++, move_cell (&tmpcell, RIGHT, 1))
	{
	  if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	    freespace--;
	}
    }
  dir[DOWN]->score += freespace * weight;
}

/* score by snake presence beetween from and to */
static void
score_by_snake_presence (snake * s, fruit * frt, scored_dir ** dir)
{
  int snakefound;
  cell *from;
  cell *to;
  cell tmpcell;
  float weight = SNAKE_PRESENCE_WEIGHT (s);

  from = &(s->head->cell);
  to = &(frt->cell);

  /* right */
  snakefound = 0;
  memcpy (&tmpcell, from, sizeof (cell));
  do
    {
      move_cell (&tmpcell, RIGHT, 1);
      if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	{
	  snakefound = 1;
	  break;
	}
    }
  while (tmpcell.x != from->x);
  if (!snakefound)
    dir[RIGHT]->score += MAXSCORE * weight;

  /* left */
  snakefound = 0;
  memcpy (&tmpcell, from, sizeof (cell));
  do
    {
      move_cell (&tmpcell, LEFT, 1);
      if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	{
	  snakefound = 1;
	  break;
	}
    }
  while (tmpcell.x != from->x);
  if (!snakefound)
    dir[LEFT]->score += MAXSCORE * weight;

  /* up */
  snakefound = 0;
  memcpy (&tmpcell, from, sizeof (cell));
  do
    {
      move_cell (&tmpcell, UP, 1);
      if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	{
	  snakefound = 1;
	  break;
	}
    }
  while (tmpcell.y != from->y);
  if (!snakefound)
    dir[UP]->score += MAXSCORE * weight;

  /* down */
  snakefound = 0;
  memcpy (&tmpcell, from, sizeof (cell));
  do
    {
      move_cell (&tmpcell, DOWN, 1);
      if (IS_SNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	{
	  snakefound = 1;
	  break;
	}
    }
  while (tmpcell.y != from->y);
  if (!snakefound)
    dir[DOWN]->score += MAXSCORE * weight;

}

/* score by fruit distace */
static void
score_by_fruit_distance (snake * s, fruit * frt, scored_dir ** dir)
{
  float maxdist, dist;
  float weight = FRUIT_DISTANCE_WEIGHT (s);
  cell *from;
  cell *to;

  from = &(s->head->cell);
  to = &(frt->cell);

  /* vertical distances */
  maxdist = ARENA_HEIGHT;

  dist = distance (from, to, DOWN);
  if (dist)
    {
      dir[DOWN]->score += (MAXSCORE - ((dist / maxdist) * MAXSCORE)) * weight;
    }

  dist = distance (from, to, UP);
  if (dist)
    {
      dir[UP]->score += (MAXSCORE - ((dist / maxdist) * MAXSCORE)) * weight;
    }

  /* horizontal distances */
  maxdist = ARENA_WIDTH;

  dist = distance (from, to, RIGHT);
  if (dist)
    {
      dir[RIGHT]->score +=
	(MAXSCORE - ((dist / maxdist) * MAXSCORE)) * weight;
    }

  dist = distance (from, to, LEFT);
  if (dist)
    {
      dir[LEFT]->score += (MAXSCORE - ((dist / maxdist) * MAXSCORE)) * weight;
    }
}

/* score by dead snake proximity */
static void
score_by_deadsnake_proximity (snake * s, scored_dir ** dir)
{
  float weight = DEADSNAKE_PROXIMITY_WEIGHT (s);
  cell tmpcell;
  int i;

  for (i = 0; i < 4; i++)
    {
      memcpy (&tmpcell, &(s->head->cell), sizeof (cell));
      move_cell (&tmpcell, i, 1);
      if (IS_EXSNAKE (ARENAGET (tmpcell.y, tmpcell.x)))
	dir[i]->score += MAXSCORE * weight;
    }
}


#ifdef DEBUG
static void
dump_scores (snake * s, scored_dir ** dirs)
{
  char buf[50], buf2[15];
  int i = 0;

  sprintf (buf, "%c): ", (char) s->look & A_CHARTEXT);
  for (i = 0; i < 4; i++)
    {
      snprintf (buf2, 15, "%c: %4.2f ",
		Direction[dirs[i]->dir], dirs[i]->score);
      strncat (buf, buf2, 15);
    }

  fprintf (stderr, "%s\n", buf);
  fflush (stderr);
}
#endif

static void
sort_direction_by_score (snake * s, scored_dir ** sdir)
{
  int changes, i;
  scored_dir *ptr;

  do
    {
      changes = 0;
      for (i = 0; i < 3; i++)
	{
	  if (sdir[i]->score < sdir[i + 1]->score)
	    {
	      changes++;
	      ptr = sdir[i];
	      sdir[i] = sdir[i + 1];
	      sdir[i + 1] = ptr;
	    }
	}
    }
  while (changes);
}


/* make a new brain */
static brain *
make_brain (float w_snake_distance, float w_fruit_distance, float w_random,
	    float w_inertia, float w_snake_presence, float w_free_space,
	    float w_deadsnake)
{
  brain *newbrain;

  newbrain = (brain *) malloc (sizeof (brain));

  newbrain->snake_distance_weight = w_snake_distance;
  newbrain->fruit_distance_weight = w_fruit_distance;
  newbrain->random_weight = w_random;
  newbrain->inertia_weight = w_inertia;
  newbrain->snake_presence_weight = w_snake_presence;
  newbrain->free_space_weight = w_free_space;
  newbrain->deadsnake_proximity_weight = w_deadsnake;

  return newbrain;
}
